from itertools import count
import logging

import networkx

import ailment
from claripy.utils.orderedset import OrderedSet

from ...utils.graph import dfs_back_edges, subgraph_between_nodes, dominates, shallow_reverse
from .. import Analysis, register_analysis
from .utils import replace_last_statement
from .structurer_nodes import MultiNode, ConditionNode
from .graph_region import GraphRegion
from .condition_processor import ConditionProcessor

l = logging.getLogger(name=__name__)


# an ever-incrementing counter
CONDITIONNODE_ADDR = count(0xff000000)


class RegionIdentifier(Analysis):
    """
    Identifies regions within a function.
    """
    def __init__(self, func, cond_proc=None, graph=None):
        self.function = func
        self.cond_proc = cond_proc if cond_proc is not None else ConditionProcessor()
        self._graph = graph if graph is not None else self.function.graph

        self.region = None
        self._start_node = None
        self._loop_headers = None

        self._analyze()

    @staticmethod
    def slice_graph(graph, node, frontier, include_frontier=False):
        """
        Generate a slice of the graph from the head node to the given frontier.

        :param networkx.DiGraph graph: The graph to work on.
        :param node: The starting node in the graph.
        :param frontier: A list of frontier nodes.
        :param bool include_frontier: Whether the frontier nodes are included in the slice or not.
        :return: A subgraph.
        :rtype: networkx.DiGraph
        """

        subgraph = subgraph_between_nodes(graph, node, frontier, include_frontier=include_frontier)
        if not list(subgraph.nodes):
            # HACK: FIXME: for infinite loop nodes, this would return an empty set, so we include the loop body itself
            # Make sure this makes sense (EDG thinks it does)
            if (node, node) in graph.edges:
                subgraph.add_edge(node, node)
        return subgraph

    def _analyze(self):

        # make a copy of the graph
        graph = networkx.DiGraph(self._graph)

        # preprocess: make it a super graph
        self._make_supergraph(graph)

        self._start_node = self._get_start_node(graph)

        # preprocess: find loop headers
        self._loop_headers = self._find_loop_headers(graph)

        self.region = self._make_regions(graph)

    def _get_start_node(self, graph: networkx.DiGraph):
        try:
            return next(n for n in graph.nodes() if graph.in_degree(n) == 0)
        except StopIteration:
            pass

        try:
            return next(n for n in graph.nodes() if n.addr == self.function.addr)
        except StopIteration:
            raise RuntimeError("Cannot find the start node from the graph!")

    def _test_reducibility(self):

        # make a copy of the graph
        graph = networkx.DiGraph(self._graph)

        # preprocess: make it a super graph
        self._make_supergraph(graph)

        while True:

            changed = False

            # find a node with a back-edge, remove the edge (deleting the loop), and replace it with a MultiNode
            changed |= self._remove_self_loop(graph)

            # find a node that has only one predecessor, and merge it with its predecessor (replace them with a
            # MultiNode)
            changed |= self._merge_single_entry_node(graph)

            if not changed:
                # a fixed-point is reached
                break

        # Flow graph reducibility, Hecht and Ullman
        if len(graph.nodes) == 1:
            return True

        return False

    def _make_supergraph(self, graph: networkx.DiGraph):

        while True:
            for src, dst, data in graph.edges(data=True):
                type_ = data.get('type', None)
                if type_ == 'fake_return':
                    if len(list(graph.successors(src))) == 1 and len(list(graph.predecessors(dst))) == 1:
                        self._merge_nodes(graph, src, dst, force_multinode=True)
                        break
                elif type_ == 'call':
                    graph.remove_node(dst)
                    break
            else:
                break

    def _find_loop_headers(self, graph: networkx.DiGraph):
        return OrderedSet(sorted((t for _,t in dfs_back_edges(graph, self._start_node)), key=lambda x: x.addr))

    def _find_initial_loop_nodes(self, graph: networkx.DiGraph, head):
        # TODO optimize
        latching_nodes = { s for s,t in dfs_back_edges(graph, self._start_node) if t == head }
        loop_subgraph = self.slice_graph(graph, head, latching_nodes, include_frontier=True)
        nodes = set(loop_subgraph.nodes())
        return nodes

    def _refine_loop(self, graph: networkx.DiGraph, head, initial_loop_nodes, initial_exit_nodes):
        refined_loop_nodes = initial_loop_nodes.copy()
        refined_exit_nodes = initial_exit_nodes.copy()

        idom = networkx.immediate_dominators(graph, self._start_node)

        new_exit_nodes = refined_exit_nodes
        while len(refined_exit_nodes) > 1 and new_exit_nodes:
            new_exit_nodes = set()
            for n in list(refined_exit_nodes):
                if all(pred in refined_loop_nodes for pred in graph.predecessors(n)) and dominates(idom, head, n):
                    refined_loop_nodes.add(n)
                    refined_exit_nodes.remove(n)
                    for u in (set(graph.successors(n)) - refined_loop_nodes):
                        new_exit_nodes.add(u)
            refined_exit_nodes |= new_exit_nodes

        refined_loop_nodes = refined_loop_nodes - refined_exit_nodes

        return refined_loop_nodes, refined_exit_nodes

    def _remove_self_loop(self, graph: networkx.DiGraph):

        r = False

        while True:
            for node in graph.nodes():
                if node in graph[node]:
                    # found a self loop
                    self._remove_node(graph, node)
                    r = True
                    break
            else:
                break

        return r

    def _merge_single_entry_node(self, graph: networkx.DiGraph):

        r = False

        while True:
            for node in networkx.dfs_postorder_nodes(graph):
                preds = graph.predecessors(node)
                if len(preds) == 1:
                    # merge the two nodes
                    self._absorb_node(graph, preds[0], node)
                    r = True
                    break
            else:
                break

        return r

    def _make_regions(self, graph: networkx.DiGraph):

        structured_loop_headers = set()
        new_regions = [ ]

        # FIXME: _get_start_node() will fail if the graph is just a loop

        # Find all loops
        while True:
            restart = False

            self._start_node = self._get_start_node(graph)

            # Start from loops
            for node in self._loop_headers:
                if node in structured_loop_headers:
                    continue
                region = self._make_cyclic_region(node, graph)
                if region is not None:
                    l.debug("Structured a loop region %r.", region)
                    new_regions.append(region)
                    structured_loop_headers.add(node)
                    restart = True
                    break

            if restart:
                continue

            break

        new_regions.append(GraphRegion(self._get_start_node(graph), graph, None, None, False))

        l.debug("Identified %d loop regions.", len(structured_loop_headers))
        l.debug("No more loops left. Start structuring acyclic regions.")
        # No more loops left. Structure acyclic regions.
        while new_regions:
            region = new_regions.pop(0)
            head = region.head
            subgraph = region.graph

            failed_region_attempts = set()
            while self._make_acyclic_region(head, subgraph, region.graph_with_successors, failed_region_attempts,
                                            region.cyclic):
                if head not in subgraph:
                    # update head
                    head = next(iter(n for n in subgraph.nodes() if n.addr == head.addr))

            head = next(iter(n for n in subgraph.nodes() if n.addr == head.addr))
            region.head = head

        if len(graph.nodes()) == 1 and isinstance(list(graph.nodes())[0], GraphRegion):
            return list(graph.nodes())[0]
        # create a large graph region
        new_head = self._get_start_node(graph)
        region = GraphRegion(new_head, graph, None, None, False)
        return region

    #
    # Cyclic regions
    #

    def _make_cyclic_region(self, head, graph: networkx.DiGraph):

        l.debug("Found cyclic region at %#08x", head.addr)
        initial_loop_nodes = self._find_initial_loop_nodes(graph, head)
        l.debug("Initial loop nodes %s", self._dbg_block_list(initial_loop_nodes))

        # Make sure there is no other loop contained in the current loop
        if {n for n in initial_loop_nodes if n.addr != head.addr}.intersection(self._loop_headers):
            return None

        normal_entries = {n for n in graph.predecessors(head) if n not in initial_loop_nodes}
        abnormal_entries = set()
        for n in initial_loop_nodes:
            if n == head:
                continue
            preds = set(graph.predecessors(n))
            abnormal_entries |= (preds - initial_loop_nodes)
        l.debug("Normal entries %s", self._dbg_block_list(normal_entries))
        l.debug("Abnormal entries %s", self._dbg_block_list(abnormal_entries))

        initial_exit_nodes = set()
        for n in initial_loop_nodes:
            succs = set(graph.successors(n))
            initial_exit_nodes |= (succs - initial_loop_nodes)

        l.debug("Initial exit nodes %s", self._dbg_block_list(initial_exit_nodes))

        refined_loop_nodes, refined_exit_nodes = self._refine_loop(graph, head, initial_loop_nodes,
                                                                   initial_exit_nodes)
        l.debug("Refined loop nodes %s", self._dbg_block_list(refined_loop_nodes))
        l.debug("Refined exit nodes %s", self._dbg_block_list(refined_exit_nodes))

        if len(refined_exit_nodes) > 1:
            # self._get_start_node(graph)
            node_post_order = list(networkx.dfs_postorder_nodes(graph, head))
            sorted_exit_nodes = sorted(list(refined_exit_nodes), key=node_post_order.index)
            normal_exit_node = sorted_exit_nodes[0]
            abnormal_exit_nodes = set(sorted_exit_nodes[1:])
        else:
            normal_exit_node = next(iter(refined_exit_nodes)) if len(refined_exit_nodes) > 0 else None
            abnormal_exit_nodes = set()

        region = self._abstract_cyclic_region(graph, refined_loop_nodes, head, normal_entries, abnormal_entries,
                                              normal_exit_node, abnormal_exit_nodes)
        if len(region.successors) > 1:
            # multi-successor region. refinement is required
            self._refine_loop_successors(region, graph)

        return region

    def _refine_loop_successors(self, region, graph: networkx.DiGraph):
        """
        If there are multiple successors of a loop, convert them into conditional gotos. Eventually there should be
        only one loop successor.

        :param GraphRegion region:      The cyclic region to refine.
        :param networkx.DiGraph graph:  The current graph that is being structured.
        :return:                        None
        """
        if len(region.successors) <= 1:
            return

        # recover reaching conditions
        self.cond_proc.recover_reaching_conditions(region, with_successors=True)

        successors = list(region.successors)

        condnode_addr = next(CONDITIONNODE_ADDR)
        # create a new successor
        cond = ConditionNode(
            condnode_addr,
            None,
            self.cond_proc.reaching_conditions[successors[0]],
            successors[0],
            false_node=None,
        )
        for succ in successors[1:]:
            cond = ConditionNode(condnode_addr,
                                 None,
                                 self.cond_proc.reaching_conditions[succ],
                                 succ,
                                 false_node=cond,
                                 )

        g = region.graph_with_successors

        # modify region in place
        region.successors = {cond}
        for succ in successors:
            for src, _, data in list(g.in_edges(succ, data=True)):
                removed_edges = [ ]
                for src2src, _, data_ in list(g.in_edges(src, data=True)):
                    removed_edges.append((src2src, src, data_))
                    g.remove_edge(src2src, src)
                g.remove_edge(src, succ)

                # TODO: rewrite the conditional jumps in src so that it goes to cond-node instead.

                # modify the last statement of src so that it jumps to cond
                replaced_any_stmt = False
                last_stmts = self.cond_proc.get_last_statements(src)
                for last_stmt in last_stmts:
                    if isinstance(last_stmt, ailment.Stmt.ConditionalJump):
                        if last_stmt.true_target.value == succ.addr:
                            new_last_stmt = ailment.Stmt.ConditionalJump(
                                last_stmt.idx,
                                last_stmt.condition,
                                ailment.Expr.Const(None, None, condnode_addr, self.project.arch.bits),
                                last_stmt.false_target,
                                ins_addr=last_stmt.ins_addr,
                            )
                        elif last_stmt.false_target.value == succ.addr:
                            new_last_stmt = ailment.Stmt.ConditionalJump(
                                last_stmt.idx,
                                last_stmt.condition,
                                last_stmt.true_target,
                                ailment.Expr.Const(None, None, condnode_addr, self.project.arch.bits),
                                ins_addr=last_stmt.ins_addr,
                            )
                        else:
                            # none of the two branches is jumping out of the loop
                            continue
                    else:
                        new_last_stmt = ailment.Stmt.Jump(
                            last_stmt.idx,
                            ailment.Expr.Const(None, None, condnode_addr, self.project.arch.bits),
                            ins_addr=last_stmt.ins_addr,
                        )
                    replace_last_statement(src, last_stmt, new_last_stmt)
                    replaced_any_stmt = True
                if not replaced_any_stmt:
                    l.warning("No statement was replaced. Is there anything wrong?")
                    # raise Exception()

                # add src back
                for src2src, _, data_ in removed_edges:
                    g.add_edge(src2src, src, **data_)

                g.add_edge(src, cond, **data)

        # modify graph
        graph.add_edge(region, cond)
        for succ in successors:
            edge_data = graph.get_edge_data(region, succ)
            graph.remove_edge(region, succ)
            graph.add_edge(cond, succ, **edge_data)

    #
    # Acyclic regions
    #

    def _make_acyclic_region(self, head, graph: networkx.DiGraph, secondary_graph, failed_region_attempts, cyclic):
        # pre-processing

        # we need to create a copy of the original graph if
        # - there are in edges to the head node, or
        # - there are more than one end nodes

        head_inedges = list(graph.in_edges(head))
        if head_inedges:
            # we need a copy of the graph to remove edges coming into the head
            graph_copy = networkx.DiGraph(graph)
            # remove any in-edge to the head node
            for src, _ in head_inedges:
                graph_copy.remove_edge(src, head)
        else:
            graph_copy = graph

        endnodes = [node for node in graph_copy.nodes() if graph_copy.out_degree(node) == 0]
        if len(endnodes) == 0:
            # sanity check: there should be at least one end node
            l.critical("No end node is found in a supposedly acyclic graph. Is it really acyclic?")
            return False

        if len(endnodes) > 1:
            # we need a copy of the graph!
            graph_copy = networkx.DiGraph(graph_copy)

            # if this graph has multiple end nodes: create a single end node
            dummy_endnode = None
            if len(endnodes) > 1:
                dummy_endnode = "DUMMY_ENDNODE"
                for endnode in endnodes:
                    graph_copy.add_edge(endnode, dummy_endnode)
                endnodes = [ dummy_endnode ]
        else:
            dummy_endnode = None

        # compute dominator tree
        doms = networkx.immediate_dominators(graph_copy, head)

        # compute post-dominator tree
        inverted_graph = shallow_reverse(graph_copy)
        postdoms = networkx.immediate_dominators(inverted_graph, endnodes[0])

        # dominance frontiers
        df = networkx.algorithms.dominance_frontiers(graph_copy, head)

        # visit the nodes in post-order
        for node in networkx.dfs_postorder_nodes(graph_copy, source=head):
            if node is dummy_endnode:
                # skip the dummy endnode
                continue
            if cyclic and node is head:
                continue

            out_degree = graph_copy.out_degree[node]
            if out_degree == 0:
                # the root element of the region hierarchy should always be a GraphRegion,
                # so we transform it into one, if necessary
                if graph_copy.in_degree(node) == 0 and not isinstance(node, GraphRegion):
                    subgraph = networkx.DiGraph()
                    subgraph.add_node(node)
                    self._abstract_acyclic_region(graph, GraphRegion(node, subgraph, None, None, False), [],
                                                  secondary_graph=secondary_graph)
                continue

            # test if this node is an entry to a single-entry, single-successor region
            levels = 0
            postdom_node = postdoms.get(node, None)
            while postdom_node is not None:
                if (node, postdom_node) not in failed_region_attempts:
                    if self._check_region(graph_copy, node, postdom_node, doms, df):
                        frontier = [ postdom_node ]
                        region = self._compute_region(graph_copy, node, frontier, dummy_endnode=dummy_endnode)
                        if region is not None:
                            # l.debug("Walked back %d levels in postdom tree.", levels)
                            l.debug("Node %r, frontier %r.", node, frontier)
                            # l.debug("Identified an acyclic region %s.", self._dbg_block_list(region.graph.nodes()))
                            self._abstract_acyclic_region(graph, region, frontier, dummy_endnode=dummy_endnode,
                                                          secondary_graph=secondary_graph)
                            # assert dummy_endnode not in graph
                            return True

                failed_region_attempts.add((node, postdom_node))
                if not dominates(doms, node, postdom_node):
                    break
                if postdom_node is postdoms.get(postdom_node, None):
                    break
                postdom_node = postdoms.get(postdom_node, None)
                levels += 1
            # l.debug("Walked back %d levels in postdom tree and did not find anything for %r. Next.", levels, node)

        return False

    @staticmethod
    def _check_region(graph, start_node, end_node, doms, df):
        """

        :param graph:
        :param start_node:
        :param end_node:
        :param doms:
        :param df:
        :return:
        """

        # if the exit node is the header of a loop that contains the start node, the dominance frontier should only
        # contain the exit node.
        if not dominates(doms, start_node, end_node):
            frontier = df.get(start_node, set())
            for node in frontier:
                if node is not start_node and node is not end_node:
                    return False

        # no edges should enter the region.
        for node in df.get(end_node, set()):
            if dominates(doms, start_node, node) and node is not end_node:
                return False

        # no edges should leave the region.
        for node in df.get(start_node, set()):
            if node is start_node or node is end_node:
                continue
            if node not in df.get(end_node, set()):
                return False
            for pred in graph.predecessors(node):
                if dominates(doms, start_node, pred) and not dominates(doms, end_node, pred):
                    return False

        return True

    @staticmethod
    def _compute_region(graph, node, frontier, include_frontier=False, dummy_endnode=None):

        subgraph = networkx.DiGraph()
        frontier_edges = [ ]
        queue = [ node ]
        traversed = set()

        while queue:
            node_ = queue.pop()
            if node_ in frontier:
                continue
            traversed.add(node_)
            subgraph.add_node(node_)

            for succ in graph.successors(node_):
                edge_data = graph.get_edge_data(node_, succ)

                if node_ in frontier and succ in traversed:
                    if include_frontier:
                        # if frontier nodes are included, do not keep traversing their successors
                        # however, if it has an edge to an already traversed node, we should add that edge
                        subgraph.add_edge(node_, succ, **edge_data)
                    else:
                        frontier_edges.append((node_, succ, edge_data))
                    continue

                if succ is dummy_endnode:
                    continue

                if succ in frontier:
                    if not include_frontier:
                        # skip all frontier nodes
                        frontier_edges.append((node_, succ, edge_data))
                        continue
                subgraph.add_edge(node_, succ, **edge_data)
                if succ in traversed:
                    continue
                queue.append(succ)

        if dummy_endnode is not None:
            frontier = { n for n in frontier if n is not dummy_endnode }

        if subgraph.number_of_nodes() > 1:
            subgraph_with_frontier = networkx.DiGraph(subgraph)
            for src, dst, edge_data in frontier_edges:
                if dst is not dummy_endnode:
                    subgraph_with_frontier.add_edge(src, dst, **edge_data)
            # assert dummy_endnode not in frontier
            # assert dummy_endnode not in subgraph_with_frontier
            return GraphRegion(node, subgraph, frontier, subgraph_with_frontier, False)
        else:
            return None

    def _abstract_acyclic_region(self,
                                 graph: networkx.DiGraph,
                                 region,
                                 frontier,
                                 dummy_endnode=None,
                                 secondary_graph=None):

        in_edges = self._region_in_edges(graph, region, data=True)
        out_edges = self._region_out_edges(graph, region, data=True)

        nodes_set = set()
        for node_ in list(region.graph.nodes()):
            nodes_set.add(node_)
            if node_ is not dummy_endnode:
                graph.remove_node(node_)

        graph.add_node(region)

        for src, _, data in in_edges:
            if src not in nodes_set:
                graph.add_edge(src, region, **data)

        for _, dst, data in out_edges:
            if dst not in nodes_set:
                graph.add_edge(region, dst, **data)

        if frontier:
            for frontier_node in frontier:
                if frontier_node is not dummy_endnode:
                    graph.add_edge(region, frontier_node)

        if secondary_graph is not None:
            self._abstract_acyclic_region(secondary_graph, region, { })

    @staticmethod
    def _abstract_cyclic_region(graph: networkx.DiGraph, loop_nodes, head, normal_entries, abnormal_entries,
                                normal_exit_node,
                                abnormal_exit_nodes):
        region = GraphRegion(head, None, None, None, True)

        subgraph = networkx.DiGraph()
        region_outedges = [ ]

        graph.add_node(region)
        for node in loop_nodes:
            subgraph.add_node(node)
            in_edges = list(graph.in_edges(node, data=True))
            out_edges = list(graph.out_edges(node, data=True))

            for src, dst, data in in_edges:
                if src in normal_entries:
                    graph.add_edge(src, region, **data)
                elif src in abnormal_entries:
                    data['region_dst_node'] = dst
                    graph.add_edge(src, region, **data)
                elif src in loop_nodes:
                    subgraph.add_edge(src, dst, **data)
                elif src is region:
                    subgraph.add_edge(head, dst, **data)
                else:
                    assert 0

            for src, dst, data in out_edges:
                if dst in loop_nodes:
                    subgraph.add_edge(src, dst, **data)
                elif dst is region:
                    subgraph.add_edge(src, head, **data)
                elif dst is normal_exit_node:
                    region_outedges.append((node, dst))
                    graph.add_edge(region, dst, **data)
                elif dst in abnormal_exit_nodes:
                    region_outedges.append((node, dst))
                    # data['region_src_node'] = src
                    graph.add_edge(region, dst, **data)
                else:
                    assert 0

        subgraph_with_exits = networkx.DiGraph(subgraph)
        for src, dst in region_outedges:
            subgraph_with_exits.add_edge(src, dst)
        region.graph = subgraph
        region.graph_with_successors = subgraph_with_exits
        if normal_exit_node is not None:
            region.successors = [normal_exit_node]
        else:
            region.successors = [ ]
        region.successors += list(abnormal_exit_nodes)

        for node in loop_nodes:
            graph.remove_node(node)

        return region

    @staticmethod
    def _region_in_edges(graph, region, data=False):

        return list(graph.in_edges(region.head, data=data))

    @staticmethod
    def _region_out_edges(graph, region, data=False):

        out_edges = [ ]
        for node in region.graph.nodes():
            out_ = graph.out_edges(node, data=data)
            for _, dst, data_ in out_:
                if dst in region.graph:
                    continue
                out_edges.append((region, dst, data_))
        return out_edges

    def _remove_node(self, graph: networkx.DiGraph, node):  # pylint:disable=no-self-use

        in_edges = [ (src, dst, data) for (src, dst, data) in graph.in_edges(node, data=True) if not src is node ]
        out_edges = [ (src, dst, data) for (src, dst, data) in graph.out_edges(node, data=True) if not dst is node ]

        if len(in_edges) <= 1 and len(out_edges) <= 1:
            # it forms a region by itself :-)
            new_node = None

        else:
            new_node = MultiNode([ node ])

        graph.remove_node(node)

        if new_node is not None:
            for src, _, data in in_edges:
                graph.add_edge(src, new_node, **data)

            for _, dst, data in out_edges:
                graph.add_edge(new_node, dst, **data)

    def _merge_nodes(self, graph: networkx.DiGraph, node_a, node_b, force_multinode=False):  # pylint:disable=no-self-use

        in_edges = list(graph.in_edges(node_a, data=True))
        out_edges = list(graph.out_edges(node_b, data=True))

        if not force_multinode and len(in_edges) <= 1 and len(out_edges) <= 1:
            # it forms a region by itself :-)
            new_node = None

        else:
            new_node = MultiNode([ node_a, node_b ])

        graph.remove_node(node_a)
        graph.remove_node(node_b)

        if new_node is not None:
            graph.add_node(new_node)

            for src, _, data in in_edges:
                if src is node_b:
                    src = new_node
                graph.add_edge(src, new_node, **data)

            for _, dst, data in out_edges:
                if dst is node_a:
                    dst = new_node
                graph.add_edge(new_node, dst, **data)

        assert not node_a in graph
        assert not node_b in graph

    def _absorb_node(self, graph: networkx.DiGraph, node_mommy, node_kiddie, force_multinode=False):  # pylint:disable=no-self-use

        in_edges_mommy = graph.in_edges(node_mommy, data=True)
        out_edges_mommy = graph.out_edges(node_mommy, data=True)
        out_edges_kiddie = graph.out_edges(node_kiddie, data=True)

        if not force_multinode and len(in_edges_mommy) <= 1 and len(out_edges_kiddie) <= 1:
            # it forms a region by itself :-)
            new_node = None

        else:
            new_node = MultiNode([node_mommy, node_kiddie])

        graph.remove_node(node_mommy)
        graph.remove_node(node_kiddie)

        if new_node is not None:
            graph.add_node(new_node)

            for src, _, data in in_edges_mommy:
                if src == node_kiddie:
                    src = new_node
                graph.add_edge(src, new_node, **data)

            for _, dst, data in out_edges_mommy:
                if dst == node_kiddie:
                    continue
                if dst == node_mommy:
                    dst = new_node
                graph.add_edge(new_node, dst, **data)

            for _, dst, data in out_edges_kiddie:
                if dst == node_mommy:
                    dst = new_node
                graph.add_edge(new_node, dst, **data)

        assert not node_mommy in graph
        assert not node_kiddie in graph

    @staticmethod
    def _dbg_block_list(blocks):
        return [(hex(b.addr) if hasattr(b, 'addr') else repr(b)) for b in blocks]


register_analysis(RegionIdentifier, 'RegionIdentifier')



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("msdrm.dll")
prototypes = \
    {
        # 
        'DRMSetGlobalOptions': SimTypeFunction([SimTypeInt(signed=False, label="DRMGLOBALOPTIONS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["eGlobalOptions", "pvdata", "dwlen"]),
        # 
        'DRMGetClientVersion': SimTypeFunction([SimTypePointer(SimStruct({"uStructVersion": SimTypeInt(signed=False, label="UInt32"), "dwVersion": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "wszHierarchy": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "wszProductId": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "wszProductDescription": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256)}, name="DRM_CLIENT_VERSION_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDRMClientVersionInfo"]),
        # 
        'DRMInitEnvironment': SimTypeFunction([SimTypeInt(signed=False, label="DRMSECURITYPROVIDERTYPE"), SimTypeInt(signed=False, label="DRMSPECTYPE"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["eSecurityProviderType", "eSpecification", "wszSecurityProvider", "wszManifestCredentials", "wszMachineCredentials", "phEnv", "phDefaultLibrary"]),
        # 
        'DRMLoadLibrary': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DRMSPECTYPE"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "eSpecification", "wszLibraryProvider", "wszCredentials", "phLibrary"]),
        # 
        'DRMCreateEnablingPrincipal': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"uVersion": SimTypeInt(signed=False, label="UInt32"), "wszIDType": SimTypeBottom(label="PWSTR"), "wszID": SimTypeBottom(label="PWSTR")}, name="DRMID", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "hLibrary", "wszObject", "pidPrincipal", "wszCredentials", "phEnablingPrincipal"]),
        # 
        'DRMCloseHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["handle"]),
        # 
        'DRMCloseEnvironmentHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv"]),
        # 
        'DRMDuplicateHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hToCopy", "phCopy"]),
        # 
        'DRMDuplicateEnvironmentHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hToCopy", "phCopy"]),
        # 
        'DRMRegisterRevocationList': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "wszRevocationList"]),
        # 
        'DRMCheckSecurity': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "cLevel"]),
        # 
        'DRMRegisterContent': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["fRegister"]),
        # 
        'DRMEncrypt': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptoProvider", "iPosition", "cNumInBytes", "pbInData", "pcNumOutBytes", "pbOutData"]),
        # 
        'DRMDecrypt': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCryptoProvider", "iPosition", "cNumInBytes", "pbInData", "pcNumOutBytes", "pbOutData"]),
        # 
        'DRMCreateBoundLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"uVersion": SimTypeInt(signed=False, label="UInt32"), "hEnablingPrincipal": SimTypeInt(signed=False, label="UInt32"), "hSecureStore": SimTypeInt(signed=False, label="UInt32"), "wszRightsRequested": SimTypeBottom(label="PWSTR"), "wszRightsGroup": SimTypeBottom(label="PWSTR"), "idResource": SimStruct({"uVersion": SimTypeInt(signed=False, label="UInt32"), "wszIDType": SimTypeBottom(label="PWSTR"), "wszID": SimTypeBottom(label="PWSTR")}, name="DRMID", pack=False, align=None), "cAuthenticatorCount": SimTypeInt(signed=False, label="UInt32"), "rghAuthenticators": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), "wszDefaultEnablingPrincipalCredentials": SimTypeBottom(label="PWSTR"), "dwFlags": SimTypeInt(signed=False, label="UInt32")}, name="DRMBOUNDLICENSEPARAMS", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "pParams", "wszLicenseChain", "phBoundLicense", "phErrorLog"]),
        # 
        'DRMCreateEnablingBitsDecryptor': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hBoundLicense", "wszRight", "hAuxLib", "wszAuxPlug", "phDecryptor"]),
        # 
        'DRMCreateEnablingBitsEncryptor': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hBoundLicense", "wszRight", "hAuxLib", "wszAuxPlug", "phEncryptor"]),
        # 
        'DRMAttest': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="DRMATTESTTYPE"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnablingPrincipal", "wszData", "eType", "pcAttestedBlob", "wszAttestedBlob"]),
        # 
        'DRMGetTime': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DRMTIMETYPE"), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "eTimerIdType", "poTimeObject"]),
        # 
        'DRMGetInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="DRMENCODINGTYPE"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["handle", "wszAttribute", "peEncoding", "pcBuffer", "pbBuffer"]),
        # 
        'DRMGetEnvironmentInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="DRMENCODINGTYPE"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["handle", "wszAttribute", "peEncoding", "pcBuffer", "pbBuffer"]),
        # 
        'DRMGetProcAddress': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hLibrary", "wszProcName", "ppfnProcAddress"]),
        # 
        'DRMGetBoundLicenseObjectCount': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszSubObjectType", "pcSubObjects"]),
        # 
        'DRMGetBoundLicenseObject': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszSubObjectType", "iWhich", "phSubObject"]),
        # 
        'DRMGetBoundLicenseAttributeCount': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszAttribute", "pcAttributes"]),
        # 
        'DRMGetBoundLicenseAttribute': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="DRMENCODINGTYPE"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszAttribute", "iWhich", "peEncoding", "pcBuffer", "pbBuffer"]),
        # 
        'DRMCreateClientSession': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="DRM_STATUS_MSG"), SimTypeBottom(label="HRESULT"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="HRESULT"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pfnCallback", "uCallbackVersion", "wszGroupIDProviderType", "wszGroupID", "phClient"]),
        # 
        'DRMIsActivated': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"uVersion": SimTypeInt(signed=False, label="UInt32"), "wszPubKey": SimTypeBottom(label="PWSTR"), "wszURL": SimTypeBottom(label="PWSTR")}, name="DRM_ACTSERV_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hClient", "uFlags", "pActServInfo"]),
        # 
        'DRMActivate': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"uVersion": SimTypeInt(signed=False, label="UInt32"), "wszPubKey": SimTypeBottom(label="PWSTR"), "wszURL": SimTypeBottom(label="PWSTR")}, name="DRM_ACTSERV_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hClient", "uFlags", "uLangID", "pActServInfo", "pvContext", "hParentWnd"]),
        # 
        'DRMGetServiceLocation': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hClient", "uServiceType", "uServiceLocation", "wszIssuanceLicense", "puServiceURLLength", "wszServiceURL"]),
        # 
        'DRMCreateLicenseStorageSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "hDefaultLibrary", "hClient", "uFlags", "wszIssuanceLicense", "phLicenseStorage"]),
        # 
        'DRMAddLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hLicenseStorage", "uFlags", "wszLicense"]),
        # 
        'DRMAcquireAdvisories': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hLicenseStorage", "wszLicense", "wszURL", "pvContext"]),
        # 
        'DRMEnumerateLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "uFlags", "uIndex", "pfSharedFlag", "puCertificateDataLen", "wszCertificateData"]),
        # 
        'DRMAcquireLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "uFlags", "wszGroupIdentityCredential", "wszRequestedRights", "wszCustomData", "wszURL", "pvContext"]),
        # 
        'DRMDeleteLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "wszLicenseId"]),
        # 
        'DRMCloseSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession"]),
        # 
        'DRMDuplicateSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSessionIn", "phSessionOut"]),
        # 
        'DRMGetSecurityProvider': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["uFlags", "puTypeLen", "wszType", "puPathLen", "wszPath"]),
        # 
        'DRMEncode': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszAlgID", "uDataLen", "pbDecodedData", "puEncodedStringLen", "wszEncodedString"]),
        # 
        'DRMDecode': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszAlgID", "wszEncodedString", "puDecodedDataLen", "pbDecodedData"]),
        # 
        'DRMConstructCertificateChain': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cCertificates", "rgwszCertificates", "pcChain", "wszChain"]),
        # 
        'DRMParseUnboundLicense': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszCertificate", "phQueryRoot"]),
        # 
        'DRMCloseQueryHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hQuery"]),
        # 
        'DRMGetUnboundLicenseObjectCount': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszSubObjectType", "pcSubObjects"]),
        # 
        'DRMGetUnboundLicenseObject': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszSubObjectType", "iIndex", "phSubQuery"]),
        # 
        'DRMGetUnboundLicenseAttributeCount': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszAttributeType", "pcAttributes"]),
        # 
        'DRMGetUnboundLicenseAttribute': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="DRMENCODINGTYPE"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQueryRoot", "wszAttributeType", "iWhich", "peEncoding", "pcBuffer", "pbBuffer"]),
        # 
        'DRMGetCertificateChainCount': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszChain", "pcCertCount"]),
        # 
        'DRMDeconstructCertificateChain': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszChain", "iWhich", "pcCert", "wszCert"]),
        # 
        'DRMVerify': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="DRMATTESTTYPE"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszData", "pcAttestedData", "wszAttestedData", "peType", "pcPrincipal", "wszPrincipal", "pcManifest", "wszManifest"]),
        # 
        'DRMCreateUser': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszUserName", "wszUserId", "wszUserIdType", "phUser"]),
        # 
        'DRMCreateRight': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszRightName", "pstFrom", "pstUntil", "cExtendedInfo", "pwszExtendedInfoName", "pwszExtendedInfoValue", "phRight"]),
        # 
        'DRMCreateIssuanceLicense': SimTypeFunction([SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pstTimeFrom", "pstTimeUntil", "wszReferralInfoName", "wszReferralInfoURL", "hOwner", "wszIssuanceLicense", "hBoundLicense", "phIssuanceLicense"]),
        # 
        'DRMAddRightWithUser': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "hRight", "hUser"]),
        # 
        'DRMClearAllRights': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense"]),
        # 
        'DRMSetMetaData': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "wszContentId", "wszContentIdType", "wszSKUId", "wszSKUIdType", "wszContentType", "wszContentName"]),
        # 
        'DRMSetUsagePolicy': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="DRM_USAGEPOLICY_TYPE"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "eUsagePolicyType", "fDelete", "fExclusion", "wszName", "wszMinVersion", "wszMaxVersion", "wszPublicKey", "wszDigestAlgorithm", "pbDigest", "cbDigest"]),
        # 
        'DRMSetRevocationPoint': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "fDelete", "wszId", "wszIdType", "wszURL", "pstFrequency", "wszName", "wszPublicKey"]),
        # 
        'DRMSetApplicationSpecificData': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "fDelete", "wszName", "wszValue"]),
        # 
        'DRMSetNameAndDescription': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "fDelete", "lcid", "wszName", "wszDescription"]),
        # 
        'DRMSetIntervalTime': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "cDays"]),
        # 
        'DRMGetIssuanceLicenseTemplate': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "puIssuanceLicenseTemplateLength", "wszIssuanceLicenseTemplate"]),
        # 
        'DRMGetSignedIssuanceLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="DRM_STATUS_MSG"), SimTypeBottom(label="HRESULT"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="HRESULT"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "hIssuanceLicense", "uFlags", "pbSymKey", "cbSymKey", "wszSymKeyType", "wszClientLicensorCertificate", "pfnCallback", "wszURL", "pvContext"]),
        # 
        'DRMGetSignedIssuanceLicenseEx': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="DRM_STATUS_MSG"), SimTypeBottom(label="HRESULT"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="HRESULT"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "hIssuanceLicense", "uFlags", "pbSymKey", "cbSymKey", "wszSymKeyType", "pvReserved", "hEnablingPrincipal", "hBoundLicenseCLC", "pfnCallback", "pvContext"]),
        # 
        'DRMClosePubHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hPub"]),
        # 
        'DRMDuplicatePubHandle': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hPubIn", "phPubOut"]),
        # 
        'DRMGetUserInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hUser", "puUserNameLength", "wszUserName", "puUserIdLength", "wszUserId", "puUserIdTypeLength", "wszUserIdType"]),
        # 
        'DRMGetRightInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRight", "puRightNameLength", "wszRightName", "pstFrom", "pstUntil"]),
        # 
        'DRMGetRightExtendedInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRight", "uIndex", "puExtendedInfoNameLength", "wszExtendedInfoName", "puExtendedInfoValueLength", "wszExtendedInfoValue"]),
        # 
        'DRMGetUsers': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "uIndex", "phUser"]),
        # 
        'DRMGetUserRights': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "hUser", "uIndex", "phRight"]),
        # 
        'DRMGetMetaData': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "puContentIdLength", "wszContentId", "puContentIdTypeLength", "wszContentIdType", "puSKUIdLength", "wszSKUId", "puSKUIdTypeLength", "wszSKUIdType", "puContentTypeLength", "wszContentType", "puContentNameLength", "wszContentName"]),
        # 
        'DRMGetApplicationSpecificData': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "uIndex", "puNameLength", "wszName", "puValueLength", "wszValue"]),
        # 
        'DRMGetIssuanceLicenseInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "pstTimeFrom", "pstTimeUntil", "uFlags", "puDistributionPointNameLength", "wszDistributionPointName", "puDistributionPointURLLength", "wszDistributionPointURL", "phOwner", "pfOfficial"]),
        # 
        'DRMGetRevocationPoint': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimStruct({"wYear": SimTypeShort(signed=False, label="UInt16"), "wMonth": SimTypeShort(signed=False, label="UInt16"), "wDayOfWeek": SimTypeShort(signed=False, label="UInt16"), "wDay": SimTypeShort(signed=False, label="UInt16"), "wHour": SimTypeShort(signed=False, label="UInt16"), "wMinute": SimTypeShort(signed=False, label="UInt16"), "wSecond": SimTypeShort(signed=False, label="UInt16"), "wMilliseconds": SimTypeShort(signed=False, label="UInt16")}, name="SYSTEMTIME", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "puIdLength", "wszId", "puIdTypeLength", "wszIdType", "puURLLength", "wszRL", "pstFrequency", "puNameLength", "wszName", "puPublicKeyLength", "wszPublicKey"]),
        # 
        'DRMGetUsagePolicy': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="DRM_USAGEPOLICY_TYPE"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "uIndex", "peUsagePolicyType", "pfExclusion", "puNameLength", "wszName", "puMinVersionLength", "wszMinVersion", "puMaxVersionLength", "wszMaxVersion", "puPublicKeyLength", "wszPublicKey", "puDigestAlgorithmLength", "wszDigestAlgorithm", "pcbDigest", "pbDigest"]),
        # 
        'DRMGetNameAndDescription': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "uIndex", "pulcid", "puNameLength", "wszName", "puDescriptionLength", "wszDescription"]),
        # 
        'DRMGetOwnerLicense': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "puOwnerLicenseLength", "wszOwnerLicense"]),
        # 
        'DRMGetIntervalTime': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hIssuanceLicense", "pcDays"]),
        # 
        'DRMRepair': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        # 
        'DRMRegisterProtectedWindow': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hEnv", "hwnd"]),
        # 
        'DRMIsWindowProtected': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "pfProtected"]),
        # 
        'DRMAcquireIssuanceLicenseTemplate': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hClient", "uFlags", "pvReserved", "cTemplates", "pwszTemplateIds", "wszUrl", "pvContext"]),
    }

lib.set_prototypes(prototypes)



import angr
import logging

l = logging.getLogger(name=__name__)

class strnlen(angr.SimProcedure):
    def run(self, s, n, wchar=False): #pylint:disable=arguments-differ,unused-argument
        strlen = angr.SIM_PROCEDURES['libc']['strlen']

        maxlen = self.state.solver.eval_one(n)
        length = self.inline_call(strlen, s, maxlen=maxlen).ret_expr
        return length



from ailment.expression import BinaryOp, Const, Convert

from .base import PeepholeOptimizationExprBase


class RemoveRedundantShifts(PeepholeOptimizationExprBase):
    __slots__ = ()

    name = "Remove redundant bitshifts"
    expr_classes = (BinaryOp, )  # all expressions are allowed

    def optimize(self, expr: BinaryOp):

        # (expr << N) >> N  ==> Convert((M-N)->M, Convert(M->(M-N), expr))
        if expr.op in ("Shr", "Sar") and isinstance(expr.operands[1], Const):
            expr_a = expr.operands[0]
            n0 = expr.operands[1].value
            if isinstance(expr_a, BinaryOp) and expr_a.op == "Shl" and isinstance(expr_a.operands[1], Const):
                n1 = expr_a.operands[1].value
                if n0 == n1:
                    inner_expr = expr_a.operands[0]
                    conv_inner_expr = Convert(
                        None,
                        expr_a.bits,
                        expr_a.bits - n0,
                        expr.op == "Sar",  # is_signed
                        inner_expr,
                        **expr.tags,
                    )
                    conv_expr = Convert(
                        None,
                        expr_a.bits - n0,
                        expr.bits,
                        False,
                        conv_inner_expr,
                        **expr.tags,
                    )
                    return conv_expr

        return None



import angr

######################################
# std::terminate
######################################

class std__terminate(angr.SimProcedure): #pylint:disable=redefined-builtin
    #pylint:disable=arguments-differ

    NO_RET = True
    ALT_NAMES = ('std::terminate()', )

    def run(self):
        # FIXME: Call terminate handlers
        self.exit(1)



import claripy

from ...engines.light.engine import SimEngineLightMixin


class TopCheckerMixin(SimEngineLightMixin):
    def _is_top(self, expr) -> bool:
        if isinstance(expr, claripy.ast.Base):
            if "TOP" in expr.variables:
                return True
        return False

    def _top(self, size: int):
        return self.state.top(size)



from collections import defaultdict

import claripy
import functools

import logging
l = logging.getLogger(name=__name__)

from ..errors import AngrError, SimError, SimUnsupportedError, SimCCallError
from .. import sim_options
from ..engines.successors import SimSuccessors

exc_list = (AngrError, SimError, claripy.ClaripyError, TypeError, ValueError, ArithmeticError, MemoryError)

from . import ExplorationTechnique


class Oppologist(ExplorationTechnique):
    """
    The Oppologist is an exploration technique that forces uncooperative code through qemu.
    """

    def __init__(self):
        ExplorationTechnique.__init__(self)

    @staticmethod
    def _restore_state(old, new):
        new.release_plugin('unicorn')
        new.register_plugin('unicorn', old.unicorn.copy())
        new.options = old.options.copy()

    def _oppologize(self, simgr, state, pn, **kwargs):
        l.debug("... pn: %s", pn)

        pn.options.add(sim_options.UNICORN)
        pn.options.add(sim_options.UNICORN_AGGRESSIVE_CONCRETIZATION)
        pn.unicorn.max_steps = 1
        pn.unicorn.countdown_symbolic_stop = 0
        pn.unicorn.countdown_unsupported_stop = 0
        pn.unicorn.countdown_nonunicorn_blocks = 0
        pn.unicorn.countdown_stop_point = 0
        ss = simgr.successors(pn, throw=True, **kwargs)

        fixup = functools.partial(self._restore_state, state)

        l.debug("... successors: %s", ss)
        for s in ss.flat_successors + ss.unconstrained_successors + ss.unsat_successors + ss.successors:
            fixup(s)

        return ss

    @staticmethod
    def _combine_results(*results):
        all_results = defaultdict(list)

        final = SimSuccessors(results[0].addr, results[0].initial_state)
        final.description = 'Oppology'
        final.sort = 'Oppologist'

        for med in results:
            final.processed = True
            final.successors.extend(med.successors)
            final.all_successors.extend(med.all_successors)
            final.flat_successors.extend(med.flat_successors)
            final.unsat_successors.extend(med.unsat_successors)
            final.unconstrained_successors.extend(med.unsat_successors)

        return final

    def _delayed_oppology(self, simgr, state, e, **kwargs):
        ss = simgr.successors(state, num_inst=e.executed_instruction_count, **kwargs)
        need_oppologizing = [ s for s in ss.flat_successors if s.addr == e.ins_addr ]
        ss.flat_successors = [ s for s in ss.flat_successors if s.addr != e.ins_addr ]
        results = [ss]

        results.extend(map(functools.partial(self._oppologize, simgr, state, **kwargs), need_oppologizing))
        return self._combine_results(*results)

    def successors(self, simgr, state, **kwargs):
        try:
            kwargs.pop('throw', None)
            return simgr.successors(state, **kwargs)

        except (SimUnsupportedError, SimCCallError) as e:
            l.debug("Errored on path %s after %d instructions", state, e.executed_instruction_count)
            try:
                if e.executed_instruction_count:
                    return self._delayed_oppology(simgr, state, e, **kwargs)
                else:
                    return self._oppologize(simgr, state, state.copy(), **kwargs)
            except exc_list: #pylint:disable=broad-except
                l.error("Oppologizer hit an error while trying to perform repairs", exc_info=True)
                raise e
        except Exception: #pylint:disable=broad-except
            l.error("Original block hit an unsupported error", exc_info=True)
            raise



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_PIC_18_LE_24_PIC_18(ArchPcode):
    name = 'PIC-18:LE:24:PIC-18'
    pcode_arch = 'PIC-18:LE:24:PIC-18'
    description = 'PIC-18'
    bits = 24
    ip_offset = 0x0
    sp_offset = 0x4
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('sfrf60', 1, 0xf60),
        Register('sfrf61', 1, 0xf61),
        Register('sfrf62', 1, 0xf62),
        Register('sfrf63', 1, 0xf63),
        Register('sfrf64', 1, 0xf64),
        Register('sfrf65', 1, 0xf65),
        Register('sfrf66', 1, 0xf66),
        Register('sfrf67', 1, 0xf67),
        Register('sfrf68', 1, 0xf68),
        Register('sfrf69', 1, 0xf69),
        Register('sfrf6a', 1, 0xf6a),
        Register('rcsta2', 1, 0xf6b),
        Register('txsta2', 1, 0xf6c),
        Register('txreg2', 1, 0xf6d),
        Register('rcreg2', 1, 0xf6e),
        Register('spbreg2', 1, 0xf6f),
        Register('ccp5con', 1, 0xf70),
        Register('ccp5rl', 1, 0xf71),
        Register('ccpr5h', 1, 0xf72),
        Register('ccp4con', 1, 0xf73),
        Register('ccpr4l', 1, 0xf74),
        Register('ccpr4h', 1, 0xf75),
        Register('t4con', 1, 0xf76),
        Register('pr4', 1, 0xf77),
        Register('tmr4', 1, 0xf78),
        Register('sfrf79', 1, 0xf79),
        Register('sfrf7a', 1, 0xf7a),
        Register('sfrf7b', 1, 0xf7b),
        Register('sfrf7c', 1, 0xf7c),
        Register('sfrf7d', 1, 0xf7d),
        Register('sfrf7e', 1, 0xf7e),
        Register('sfrf7f', 1, 0xf7f),
        Register('porta', 1, 0xf80),
        Register('portb', 1, 0xf81),
        Register('portc', 1, 0xf82),
        Register('portd', 1, 0xf83),
        Register('porte', 1, 0xf84),
        Register('portf', 1, 0xf85),
        Register('portg', 1, 0xf86),
        Register('porth', 1, 0xf87),
        Register('portj', 1, 0xf88),
        Register('lata', 1, 0xf89),
        Register('latb', 1, 0xf8a),
        Register('latc', 1, 0xf8b),
        Register('latd', 1, 0xf8c),
        Register('late', 1, 0xf8d),
        Register('latf', 1, 0xf8e),
        Register('latg', 1, 0xf8f),
        Register('lath', 1, 0xf90),
        Register('latj', 1, 0xf91),
        Register('trisa', 1, 0xf92),
        Register('trisb', 1, 0xf93),
        Register('trisc', 1, 0xf94),
        Register('trisd', 1, 0xf95),
        Register('trise', 1, 0xf96),
        Register('trisf', 1, 0xf97),
        Register('trisg', 1, 0xf98),
        Register('trish', 1, 0xf99),
        Register('trisj', 1, 0xf9a),
        Register('sfrf9b', 1, 0xf9b),
        Register('memcon', 1, 0xf9c),
        Register('pie1', 1, 0xf9d),
        Register('pir1', 1, 0xf9e),
        Register('ipr1', 1, 0xf9f),
        Register('pie2', 1, 0xfa0),
        Register('pir2', 1, 0xfa1),
        Register('ipr2', 1, 0xfa2),
        Register('pie3', 1, 0xfa3),
        Register('pir3', 1, 0xfa4),
        Register('ipr3', 1, 0xfa5),
        Register('eecon1', 1, 0xfa6),
        Register('eecon2', 1, 0xfa7),
        Register('eedata', 1, 0xfa8),
        Register('eeadr', 1, 0xfa9),
        Register('eeadrh', 1, 0xfaa),
        Register('rcsta1', 1, 0xfab),
        Register('txsta1', 1, 0xfac),
        Register('txreg1', 1, 0xfad),
        Register('rcreg1', 1, 0xfae),
        Register('spbrg1', 1, 0xfaf),
        Register('pspcon', 1, 0xfb0),
        Register('t3con', 1, 0xfb1),
        Register('tmr3l', 1, 0xfb2),
        Register('tmr3h', 1, 0xfb3),
        Register('cmcon', 1, 0xfb4),
        Register('cvrcon', 1, 0xfb5),
        Register('sfrfb6', 1, 0xfb6),
        Register('ccp3con', 1, 0xfb7),
        Register('ccp3rl', 1, 0xfb8),
        Register('ccp3rh', 1, 0xfb9),
        Register('ccp2con', 1, 0xfba),
        Register('ccpr2', 2, 0xfbb),
        Register('ccpr2l', 1, 0xfbb),
        Register('ccpr2h', 1, 0xfbc),
        Register('ccp1con', 1, 0xfbd),
        Register('ccpr1', 2, 0xfbe),
        Register('ccpr1l', 1, 0xfbe),
        Register('ccpr1h', 1, 0xfbf),
        Register('adcon2', 1, 0xfc0),
        Register('adcon1', 1, 0xfc1),
        Register('adcon0', 1, 0xfc2),
        Register('adres', 2, 0xfc3),
        Register('adresl', 1, 0xfc3),
        Register('adresh', 1, 0xfc4),
        Register('sspcon2', 1, 0xfc5),
        Register('sspcon1', 1, 0xfc6),
        Register('sspstat', 1, 0xfc7),
        Register('sspadd', 1, 0xfc8),
        Register('sspbuf', 1, 0xfc9),
        Register('t2con', 1, 0xfca),
        Register('pr2', 1, 0xfcb),
        Register('tmr2', 1, 0xfcc),
        Register('t1con', 1, 0xfcd),
        Register('tmr1', 2, 0xfce),
        Register('tmr1l', 1, 0xfce),
        Register('tmr1h', 1, 0xfcf),
        Register('rcon', 1, 0xfd0),
        Register('wdtcon', 1, 0xfd1),
        Register('lvdcon', 1, 0xfd2),
        Register('osccon', 1, 0xfd3),
        Register('sfrfd4', 1, 0xfd4),
        Register('t0con', 1, 0xfd5),
        Register('tmr0', 2, 0xfd6),
        Register('tmr0l', 1, 0xfd6),
        Register('tmr0h', 1, 0xfd7),
        Register('status', 1, 0xfd8),
        Register('fsr2', 2, 0xfd9),
        Register('fsr2l', 1, 0xfd9),
        Register('fsr2h', 1, 0xfda),
        Register('plusw2', 1, 0xfdb),
        Register('preinc2', 1, 0xfdc),
        Register('postdec2', 1, 0xfdd),
        Register('postinc2', 1, 0xfde),
        Register('indf2', 1, 0xfdf),
        Register('bsr', 1, 0xfe0),
        Register('fsr1', 2, 0xfe1),
        Register('fsr1l', 1, 0xfe1),
        Register('fsr1h', 1, 0xfe2),
        Register('plusw1', 1, 0xfe3),
        Register('preinc1', 1, 0xfe4),
        Register('postdec1', 1, 0xfe5),
        Register('postinc1', 1, 0xfe6),
        Register('indf1', 1, 0xfe7),
        Register('wreg', 1, 0xfe8),
        Register('fsr0', 2, 0xfe9),
        Register('fsr0l', 1, 0xfe9),
        Register('fsr0h', 1, 0xfea),
        Register('plusw0', 1, 0xfeb),
        Register('preinc0', 1, 0xfec),
        Register('postdec0', 1, 0xfed),
        Register('postinc0', 1, 0xfee),
        Register('indf0', 1, 0xfef),
        Register('intcon3', 1, 0xff0),
        Register('intcon2', 1, 0xff1),
        Register('intcon', 1, 0xff2),
        Register('prod', 2, 0xff3),
        Register('prodl', 1, 0xff3),
        Register('prodh', 1, 0xff4),
        Register('tablat', 1, 0xff5),
        Register('tblptr', 3, 0xff6),
        Register('tblptrl', 1, 0xff6),
        Register('tblptrh', 1, 0xff7),
        Register('tblptru', 1, 0xff8),
        Register('pclat', 3, 0xff9),
        Register('pcl', 1, 0xff9),
        Register('pclath', 1, 0xffa),
        Register('pclatu', 1, 0xffb),
        Register('.stkptr', 1, 0xffc),
        Register('tos', 3, 0xffd),
        Register('tosl', 1, 0xffd),
        Register('tosh', 1, 0xffe),
        Register('tosu', 1, 0xfff),
        Register('pc', 3, 0x0, alias_names=('ip',)),
        Register('bad', 1, 0x3),
        Register('stkptr', 1, 0x4),
        Register('n', 1, 0x5),
        Register('ov', 1, 0x6),
        Register('z', 1, 0x7),
        Register('dc', 1, 0x8),
        Register('c', 1, 0x9),
        Register('ws', 1, 0xa),
        Register('statuss', 1, 0xb),
        Register('bsrs', 1, 0xc)
    ]

register_arch(['pic-18:le:24:pic-18'], 24, Endness.LE, ArchPcode_PIC_18_LE_24_PIC_18)



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_8051_BE_16_default(ArchPcode):
    name = '8051:BE:16:default'
    pcode_arch = '8051:BE:16:default'
    description = '8051 Microcontroller Family'
    bits = 16
    ip_offset = 0x44
    sp_offset = 0x40
    bp_offset = sp_offset
    instruction_endness = Endness.BE
    register_list = [
        Register('r0r1r2r3', 4, 0x0),
        Register('r0r1', 2, 0x0),
        Register('r0', 1, 0x0),
        Register('r1r2r3', 3, 0x1),
        Register('r2r1', 2, 0x1),
        Register('r1', 1, 0x1),
        Register('r2r3', 2, 0x2),
        Register('r2', 1, 0x2),
        Register('r3', 1, 0x3),
        Register('r4r5r6r7', 4, 0x4),
        Register('r4r5', 2, 0x4),
        Register('r4', 1, 0x4),
        Register('r5r6r7', 3, 0x5),
        Register('r5', 1, 0x5),
        Register('r6r7', 2, 0x6),
        Register('r6', 1, 0x6),
        Register('r7', 1, 0x7),
        Register('ab', 2, 0xa),
        Register('b', 1, 0xa),
        Register('acc', 1, 0xb),
        Register('sp', 1, 0x40),
        Register('pc', 2, 0x44, alias_names=('ip',)),
        Register('psw', 1, 0x48),
        Register('jumptableguard1', 1, 0x70),
        Register('jumptableguard2', 1, 0x71),
        Register('dptr', 2, 0x82),
        Register('dph', 1, 0x82),
        Register('dpl', 1, 0x83)
    ]

register_arch(['8051:be:16:default'], 16, Endness.BE, ArchPcode_8051_BE_16_default)



import ailment

from ....analyses import AnalysesHub
from ...analysis import Analysis
from ..empty_node_remover import EmptyNodeRemover
from .goto import GotoSimplifier
from .if_ import IfSimplifier
from .cascading_ifs import CascadingIfsRemover
from .ifelse import IfElseFlattener
from .loop import LoopSimplifier
from .expr_folding import ExpressionCounter, ExpressionFolder
from .cascading_cond_transformer import CascadingConditionTransformer


class RegionSimplifier(Analysis):
    """
    Simplifies a given region.
    """
    def __init__(self, func, region, variable_kb=None):
        self.func = func
        self.region = region
        self.variable_kb = variable_kb

        self.result = None

        self._simplify()

    def _simplify(self):
        """
        RegionSimplifier performs the following simplifications:
        - Remove redundant Gotos
        - Remove redundant If/If-else statements
        """

        r = self.region
        if self.variable_kb is not None:
            # Fold expressions that are only used once into their use sites
            r = self._fold_oneuse_expressions(r)
        # Remove empty nodes
        r = self._remove_empty_nodes(r)
        # Find nested if-else constructs and convert them into CascadingIfs
        r = self._transform_to_cascading_ifs(r)
        # Remove unnecessary Jump statements
        r = self._simplify_gotos(r)
        # Remove unnecessary jump or conditional jump statements if they jump to the successor right afterwards
        r = self._simplify_ifs(r)
        # Remove unnecessary else branches if the if branch will always return
        r = self._simplify_ifelses(r)
        #
        r = self._simplify_cascading_ifs(r)
        #
        r = self._simplify_loops(r)

        self.result = r

    #
    # Simplifiers
    #

    def _fold_oneuse_expressions(self, region):
        expr_counter = ExpressionCounter(region, self.variable_kb.variables[self.func.addr])

        variable_assignments = {}
        variable_uses = {}
        for var, uses in expr_counter.uses.items():
            if len(uses) == 1 \
                    and var in expr_counter.assignments \
                    and len(expr_counter.assignments[var]) == 1:

                definition, deps, loc = next(iter(expr_counter.assignments[var]))
                # make sure all variables that var depends on has only been assigned once
                fail = False
                for dep_var in deps:
                    if dep_var not in expr_counter.assignments or len(expr_counter.assignments[dep_var]) != 1:
                        fail = True
                        break
                if fail:
                    continue

                if isinstance(definition, ailment.Stmt.Call):
                    # clear the existing variable since we no longer write to this variable after expression folding
                    definition = definition.copy()
                    definition.ret_expr = definition.ret_expr.copy()
                    definition.ret_expr.variable = None
                variable_assignments[var] = definition, loc
                variable_uses[var] = next(iter(expr_counter.uses[var]))

        # replace them
        ExpressionFolder(variable_assignments, variable_uses, region)
        return region

    @staticmethod
    def _remove_empty_nodes(region):
        return EmptyNodeRemover(region).result

    @staticmethod
    def _transform_to_cascading_ifs(region):
        CascadingConditionTransformer(region)
        return region

    @staticmethod
    def _simplify_gotos(region):
        GotoSimplifier(region)
        return region

    @staticmethod
    def _simplify_ifs(region):
        IfSimplifier(region)
        return region

    def _simplify_ifelses(self, region):
        IfElseFlattener(region, self.kb.functions)
        return region

    @staticmethod
    def _simplify_cascading_ifs(region):
        CascadingIfsRemover(region)
        return region

    @staticmethod
    def _simplify_loops(region):
        LoopSimplifier(region)
        return region


AnalysesHub.register_default('RegionSimplifier', RegionSimplifier)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("davclnt.dll")
prototypes = \
    {
        # 
        'DavGetTheLockOwnerOfTheFile': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["FileName", "LockOwnerName", "LockOwnerNameLengthInBytes"]),
        # 
        'DavInvalidateCache': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["URLName"]),
        # 
        'DavCancelConnectionsToServer': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpName", "fForce"]),
        # 
        'DavRegisterAuthCallback': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"AuthBlob": SimStruct({"pBuffer": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "ulSize": SimTypeInt(signed=False, label="UInt32"), "ulType": SimTypeInt(signed=False, label="UInt32")}, name="DAV_CALLBACK_AUTH_BLOB", pack=False, align=None), "UNPBlob": SimStruct({"pszUserName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ulUserNameLength": SimTypeInt(signed=False, label="UInt32"), "pszPassword": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ulPasswordLength": SimTypeInt(signed=False, label="UInt32")}, name="DAV_CALLBACK_AUTH_UNP", pack=False, align=None), "bAuthBlobValid": SimTypeInt(signed=True, label="Int32"), "bSave": SimTypeInt(signed=True, label="Int32")}, name="DAV_CALLBACK_CRED", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="AUTHNEXTSTEP"), offset=0), SimTypePointer(SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pbuffer"]), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpwzServerName", "lpwzRemoteName", "dwAuthScheme", "dwFlags", "pCallbackCred", "NextStep", "pFreeCred"]), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["CallBack", "Version"]),
        # 
        'DavUnregisterAuthCallback': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hCallback"]),
        # 
        'NPAddConnection': SimTypeFunction([SimTypePointer(SimStruct({"dwScope": SimTypeInt(signed=False, label="NET_RESOURCE_SCOPE"), "dwType": SimTypeInt(signed=False, label="NET_RESOURCE_TYPE"), "dwDisplayType": SimTypeInt(signed=False, label="UInt32"), "dwUsage": SimTypeInt(signed=False, label="UInt32"), "lpLocalName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpRemoteName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpProvider": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="NETRESOURCEW", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpNetResource", "lpPassword", "lpUserName"]),
        # 
        'NPAddConnection3': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"dwScope": SimTypeInt(signed=False, label="NET_RESOURCE_SCOPE"), "dwType": SimTypeInt(signed=False, label="NET_RESOURCE_TYPE"), "dwDisplayType": SimTypeInt(signed=False, label="UInt32"), "dwUsage": SimTypeInt(signed=False, label="UInt32"), "lpLocalName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpRemoteName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpProvider": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="NETRESOURCEW", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="NET_USE_CONNECT_FLAGS")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hwndOwner", "lpNetResource", "lpPassword", "lpUserName", "dwFlags"]),
        # 
        'NPCancelConnection': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpName", "fForce"]),
        # 
        'NPGetConnection': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpLocalName", "lpRemoteName", "lpnBufferLen"]),
        # 
        'NPGetUniversalName': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UNC_INFO_LEVEL"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpLocalPath", "dwInfoLevel", "lpBuffer", "lpBufferSize"]),
        # 
        'NPOpenEnum': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwScope": SimTypeInt(signed=False, label="NET_RESOURCE_SCOPE"), "dwType": SimTypeInt(signed=False, label="NET_RESOURCE_TYPE"), "dwDisplayType": SimTypeInt(signed=False, label="UInt32"), "dwUsage": SimTypeInt(signed=False, label="UInt32"), "lpLocalName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpRemoteName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpProvider": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="NETRESOURCEW", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwScope", "dwType", "dwUsage", "lpNetResource", "lphEnum"]),
        # 
        'NPEnumResource': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hEnum", "lpcCount", "lpBuffer", "lpBufferSize"]),
        # 
        'NPCloseEnum': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hEnum"]),
        # 
        'NPGetCaps': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["ndex"]),
        # 
        'NPGetUser': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpName", "lpUserName", "lpnBufferLen"]),
        # 
        'NPGetResourceParent': SimTypeFunction([SimTypePointer(SimStruct({"dwScope": SimTypeInt(signed=False, label="NET_RESOURCE_SCOPE"), "dwType": SimTypeInt(signed=False, label="NET_RESOURCE_TYPE"), "dwDisplayType": SimTypeInt(signed=False, label="UInt32"), "dwUsage": SimTypeInt(signed=False, label="UInt32"), "lpLocalName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpRemoteName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpProvider": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="NETRESOURCEW", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpNetResource", "lpBuffer", "lpBufferSize"]),
        # 
        'NPGetResourceInformation': SimTypeFunction([SimTypePointer(SimStruct({"dwScope": SimTypeInt(signed=False, label="NET_RESOURCE_SCOPE"), "dwType": SimTypeInt(signed=False, label="NET_RESOURCE_TYPE"), "dwDisplayType": SimTypeInt(signed=False, label="UInt32"), "dwUsage": SimTypeInt(signed=False, label="UInt32"), "lpLocalName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpRemoteName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpProvider": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="NETRESOURCEW", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpNetResource", "lpBuffer", "lpBufferSize", "lplpSystem"]),
        # 
        'NPFormatNetworkName': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="NETWORK_NAME_FORMAT_FLAGS"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpRemoteName", "lpFormattedName", "lpnLength", "dwFlags", "dwAveCharPerLine"]),
    }

lib.set_prototypes(prototypes)



from typing import Dict, Any, TYPE_CHECKING
import json

from ...analyses.decompiler.structured_codegen import DummyStructuredCodeGenerator
from ...analyses.decompiler.decompilation_cache import DecompilationCache
from ...knowledge_plugins import StructuredCodeManager
from ..models import DbStructuredCode

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase
    from angr.angrdb.models import DbKnowledgeBase


class StructuredCodeManagerSerializer:
    """
    Serialize/unserialize a structured code manager.
    """

    @staticmethod
    def dump(session, db_kb: 'DbKnowledgeBase', code_manager: StructuredCodeManager):
        """

        :param session:
        :param db_kb:
        :param code_manager:
        :return:
        """

        # remove all existing stored structured code
        session.query(DbStructuredCode).filter_by(kb=db_kb).delete()

        for key, cache in code_manager.cached.items():
            func_addr, flavor = key

            # TODO: Cache types

            expr_comments = None
            if cache.codegen.expr_comments:
                expr_comments = json.dumps(cache.codegen.expr_comments).encode("utf-8")

            stmt_comments = None
            if cache.codegen.stmt_comments:
                stmt_comments = json.dumps(cache.codegen.stmt_comments).encode("utf-8")

            db_code = DbStructuredCode(
                kb=db_kb,
                func_addr=func_addr,
                flavor=flavor,
                expr_comments=expr_comments,
                stmt_comments=stmt_comments,
                # configuration=configuration,
            )
            session.add(db_code)

    @staticmethod
    def dict_strkey_to_intkey(d: Dict[str,Any]) -> Dict[int,Any]:
        new_d = {}

        for key, value in d.items():
            try:
                new_d[int(key)] = value
            except (ValueError, TypeError):
                continue

        return new_d

    @staticmethod
    def load(session, db_kb: 'DbKnowledgeBase', kb: 'KnowledgeBase') -> StructuredCodeManager:
        """

        :param session:
        :param db_kb:
        :param kb:
        :return:                        A loaded structured code manager
        """

        manager = StructuredCodeManager(kb)

        db_code_collection = session.query(DbStructuredCode).filter_by(kb=db_kb)

        for db_code in db_code_collection:
            if not db_code.expr_comments:
                expr_comments = None
            else:
                expr_comments = json.loads(db_code.expr_comments.decode("utf-8"))
                expr_comments = StructuredCodeManagerSerializer.dict_strkey_to_intkey(expr_comments)

            if not db_code.stmt_comments:
                stmt_comments = None
            else:
                stmt_comments = json.loads(db_code.stmt_comments.decode("utf-8"))
                stmt_comments = StructuredCodeManagerSerializer.dict_strkey_to_intkey(stmt_comments)
            configuration = None
            dummy_codegen = DummyStructuredCodeGenerator(db_code.flavor,
                                                         expr_comments=expr_comments,
                                                         stmt_comments=stmt_comments,
                                                         configuration=configuration,
                                                         )
            cache = DecompilationCache()
            cache.codegen = dummy_codegen
            manager[(db_code.func_addr, db_code.flavor)] = cache

        return manager




from ..glibc.__libc_start_main import __libc_start_main as fucker

######################################
# __uClibc_main
######################################
class __uClibc_main(fucker):
    ADDS_EXITS = True
    NO_RET = True

    # This is called "fucker" cause otherwise the double underscores cause
    # python to name-mangle and everything gets fucked.



import angr

######################################
# __errno_location
######################################

class __errno_location(angr.SimProcedure):
    def run(self):  #pylint:disable=arguments-differ
        return self.state.libc.errno_location



default_plugins = {}


class KnowledgeBasePlugin:

    def copy(self):
        raise NotImplementedError

    @staticmethod
    def register_default(name, cls):
        if name in default_plugins:
            raise Exception("%s is already set as the default for %s" % (default_plugins[name], name))
        default_plugins[name] = cls



# pylint:disable=import-outside-toplevel
from typing import Dict, Any, Union, TYPE_CHECKING

from .. import KnowledgeBasePlugin

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase
    from angr.analyses.decompiler.structured_codegen import BaseStructuredCodeGenerator
    from angr.analyses.decompiler.decompilation_cache import DecompilationCache


class StructuredCodeManager(KnowledgeBasePlugin):
    def __init__(self, kb):
        self._kb = kb  # type: KnowledgeBase
        self.cached: Dict[Any, 'DecompilationCache'] = {}

    def _normalize_key(self, item):
        if type(item) is not tuple:
            raise TypeError("Structured code can only be queried by tuples of (func, flavor)")
        if type(item[0]) is str:
            item = (self._kb.labels.lookup(item[0]), *item[1:])
        return item

    def __getitem__(self, item) -> 'DecompilationCache':
        return self.cached[self._normalize_key(item)]

    def __setitem__(self, key, value: Union['DecompilationCache','BaseStructuredCodeGenerator']):
        from ...analyses.decompiler.structured_codegen import BaseStructuredCodeGenerator
        from ...analyses.decompiler.decompilation_cache import DecompilationCache

        if isinstance(value, BaseStructuredCodeGenerator):
            cache = DecompilationCache()
            cache.codegen = value
        else:
            cache = value
        self.cached[self._normalize_key(key)] = cache

    def __contains__(self, key):
        return self._normalize_key(key) in self.cached

    def __delitem__(self, key):
        del self.cached[self._normalize_key(key)]

    def discard(self, key):
        normalized_key = self._normalize_key(key)
        if normalized_key in self.cached:
            del self.cached[normalized_key]

    def available_flavors(self, item):
        if type(item) is str:
            item = self._kb.labels.lookup(item)
        return [flavor for func, flavor in self.cached if func == item]

    def copy(self):
        raise NotImplementedError


KnowledgeBasePlugin.register_default('structured_code', StructuredCodeManager)



from typing import Dict, Any, Tuple, Iterable, Generator, Optional, TYPE_CHECKING

import claripy
import ailment

if TYPE_CHECKING:
    from ...code_location import CodeLocation


class Detail:
    """
    A companion class used together with PropValue. It describes stored information at each offset (in bytes).
    """

    __slots__ = ('size', 'expr', 'def_at')

    def __init__(self, size: int, expr: Optional[ailment.Expression], def_at: 'CodeLocation'):
        self.size = size
        self.expr = expr
        self.def_at = def_at

    def __repr__(self) -> str:
        return f"{self.size:x}: {self.expr}@{self.def_at}"


# The base value

class PropValue:
    """
    Describes immutable basic value type that is used in Propagator.
    """

    __slots__ = ('value', 'offset_and_details',)

    def __init__(self,
                 value: claripy.ast.Bits,
                 offset_and_details: Optional[Dict[int,Detail]] = None):
        self.value = value
        self.offset_and_details = offset_and_details

    @property
    def needs_details(self):
        return not bool(self.offset_and_details)

    @property
    def one_expr(self) -> Optional[ailment.Expression]:
        """
        Get the expression that starts at offset 0 and covers the entire PropValue. Returns None if there are no
        expressions or multiple expressions.
        """
        if self.offset_and_details and len(self.offset_and_details) == 1:
            if 0 in self.offset_and_details:
                detail = self.offset_and_details[0]
                if detail.size == self.value.size() // 8:
                    return detail.expr
        return None

    @property
    def one_defat(self) -> Optional['CodeLocation']:
        """
        Get the definition location of the expression that starts at offset 0 and covers the entire PropValue. Returns
        None if there are no expressions or multiple expressions.
        """
        if self.offset_and_details and len(self.offset_and_details) == 1:
            if 0 in self.offset_and_details:
                detail = self.offset_and_details[0]
                if detail.size == self.value.size() // 8:
                    return detail.def_at
        return None

    def to_label(self):
        raise NotImplementedError()

    def with_details(self, size: int, expr: ailment.Expression, def_at: 'CodeLocation') -> 'PropValue':
        return PropValue(self.value, offset_and_details={0: Detail(size, expr, def_at)})

    def all_exprs(self) -> Generator[ailment.Expression,None,None]:
        if not self.offset_and_details:
            return
        for detail in self.offset_and_details.values():
            yield detail.expr

    def non_zero_exprs(self) -> Generator[ailment.Expression,None,None]:
        if not self.offset_and_details:
            return
        for detail in self.offset_and_details.values():
            if isinstance(detail.expr, ailment.Expr.Const) and detail.expr.value == 0:
                continue
            yield detail.expr

    @staticmethod
    def chop_value(value: claripy.ast.Bits, begin_offset, end_offset) -> claripy.ast.Bits:
        chop_start = value.size() - begin_offset * 8 - 1
        chop_end = value.size() - end_offset * 8
        return value[chop_start : chop_end]

    def value_and_labels(self) -> Generator[Tuple[int,claripy.ast.Bits,int,Optional[Dict]],None,None]:
        keys = list(sorted(self.offset_and_details.keys()))
        if keys[0] != 0:
            # missing details at 0
            yield 0, self.chop_value(self.value, 0, keys[0]), keys[0], None

        end_offset = 0
        for idx, offset in enumerate(keys):
            detail = self.offset_and_details[offset]
            end_offset = offset + detail.size
            label = {
                'expr': detail.expr,
                'def_at': detail.def_at
            }
            yield offset, self.chop_value(self.value, offset, end_offset), end_offset - offset, label

            # gap detection
            if idx != len(keys) - 1:
                next_offset = keys[idx + 1]
                if end_offset != next_offset:
                    yield end_offset, self.chop_value(self.value, end_offset, next_offset), next_offset - end_offset, \
                          None

        # final gap detection
        if end_offset < self.value.size() // 8:
            yield end_offset, self.chop_value(self.value, end_offset, self.value.size() // 8), \
                self.value.size() // 8 - end_offset, None

    @staticmethod
    def from_value_and_labels(value: claripy.ast.Bits,
                              labels: Iterable[Tuple[int,int,int,Dict[str,Any]]]) -> 'PropValue':
        if not labels:
            return PropValue(value)
        offset_and_details = {}
        for offset, offset_in_expr, size, label in labels:
            expr = label['expr']
            if expr is not None:
                if offset_in_expr != 0:
                    expr = PropValue.extract_ail_expression(offset_in_expr * 8, size * 8, expr)
                elif size < expr.size:
                    expr = PropValue.extract_ail_expression(0, size * 8, expr)
                elif size > expr.size:
                    expr = PropValue.extend_ail_expression((size - expr.size) * 8, expr)
            offset_and_details[offset] = Detail(size, expr, label['def_at'])
        return PropValue(value, offset_and_details=offset_and_details)

    @staticmethod
    def from_value_and_details(value: claripy.ast.Bits,
                               size: int,
                               expr: ailment.Expression,
                               def_at: 'CodeLocation'):
        d = Detail(size, expr, def_at)
        return PropValue(value, offset_and_details={0: d})

    @staticmethod
    def extract_ail_expression(start: int, bits: int,
                               expr: Optional[ailment.Expr.Expression]) -> Optional[ailment.Expr.Expression]:
        if expr is None:
            return None

        if isinstance(expr, ailment.Expr.Const):
            mask = (1 << bits) - 1
            return ailment.Expr.Const(expr.idx, expr.variable, (expr.value >> start) & mask, bits, **expr.tags)

        if start == 0:
            return ailment.Expr.Convert(None, expr.bits, bits, False, expr)
        else:
            a = ailment.Expr.BinaryOp(None, "Shr", (expr, ailment.Expr.Const(None, None, bits, expr.bits)), False)
            return ailment.Expr.Convert(None, a.bits, bits, False, a)

    @staticmethod
    def extend_ail_expression(bits: int, expr: Optional[ailment.Expr.Expression]) -> Optional[ailment.Expr.Expression]:
        if expr is None:
            return None
        if isinstance(expr, ailment.Expr.Const):
            return ailment.Expr.Const(expr.idx, expr.variable, expr.value, bits + expr.bits, **expr.tags)
        elif isinstance(expr, ailment.Expr.Convert):
            return ailment.Expr.Convert(None, expr.from_bits, bits + expr.to_bits, False, expr.operand, **expr.tags)
        return ailment.Expr.Convert(None, expr.bits, bits + expr.bits, False, expr)



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_8085_LE_16_default(ArchPcode):
    name = '8085:LE:16:default'
    pcode_arch = '8085:LE:16:default'
    description = 'Intel 8085'
    bits = 16
    ip_offset = 0x20
    sp_offset = 0x22
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('af', 2, 0x0),
        Register('f', 1, 0x0),
        Register('a', 1, 0x1),
        Register('bc', 2, 0x2),
        Register('c', 1, 0x2),
        Register('b', 1, 0x3),
        Register('de', 2, 0x4),
        Register('e', 1, 0x4),
        Register('d', 1, 0x5),
        Register('hl', 2, 0x6),
        Register('l', 1, 0x6),
        Register('h', 1, 0x7),
        Register('af_', 2, 0x10),
        Register('a_', 1, 0x10),
        Register('f_', 1, 0x11),
        Register('bc_', 2, 0x12),
        Register('b_', 1, 0x12),
        Register('c_', 1, 0x13),
        Register('de_', 2, 0x14),
        Register('d_', 1, 0x14),
        Register('e_', 1, 0x15),
        Register('hl_', 2, 0x16),
        Register('h_', 1, 0x16),
        Register('l_', 1, 0x17),
        Register('pc', 2, 0x20, alias_names=('ip',)),
        Register('sp', 2, 0x22),
        Register('s_flag', 1, 0x30),
        Register('z_flag', 1, 0x31),
        Register('ac_flag', 1, 0x32),
        Register('p_flag', 1, 0x33),
        Register('cy_flag', 1, 0x34)
    ]

register_arch(['8085:le:16:default'], 16, Endness.LE, ArchPcode_8085_LE_16_default)



from collections import defaultdict

import claripy

from ..analyses import AnalysesHub
from ..analyses.reaching_definitions.function_handler import FunctionHandler
from ..knowledge_plugins.key_definitions.atoms import Register, MemoryLocation
from ..storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from ..knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
from . import Analysis


class PossibleObject:
    """
    This holds the address and class name of possible class instances.
    The address that it holds in mapped outside the binary so it is only valid in this analysis.
    TO DO: map the address to its uses in the registers/memory locations in the instructions
    """

    def __init__(self, size, addr, class_name=None):
        self.size = size
        # This address is only valid during RDA as we map new objects outside the already mapped region
        self.addr = addr
        self.class_name = class_name

    def __hash__(self):
        return hash((self.addr, self.size, self.class_name))

    def __eq__(self, other):
        return (
            self.size == other.size
            and self.addr == other.size
            and self.class_name == other.class_name
        )


class NewFunctionHandler(FunctionHandler):
    """
    This handles calls to the function new(), by recording the size parameter passed to it and also assigns a new
     address outside the mapped binary to the newly created space(possible object).

     It also tracks if the function called right after new() is passed the same 'this' pointer and is a constructor,
     if so we mark it as an instance of the class the constructor belongs to.(only for non stripped binaries)
    """

    def __init__(self, max_addr=None, new_func_addr=None, project=None):
        self.max_addr = max_addr

        # this is a map between an object addr outside the mapped binary and PossibleObject instance
        self.possible_objects_dict = {}

        # address of the new() function
        self.new_func_addr = new_func_addr
        self.project = project

    def hook(self, analysis):
        return self

    def handle_local_function(
        self,
        state,
        function_address,
        call_stack,
        maximum_local_call_depth,
        visited_blocks,
        dep_graph,
        src_ins_addr=None,
        codeloc=None,
    ):

        if function_address == self.new_func_addr:
            word_size = self.project.arch.bits // self.project.arch.byte_width
            # check if this is a call to new()
            # reading from rdi for the size argument passed to new()
            cc = self.project.kb.functions[function_address].calling_convention
            if cc is not None:
                size_arg_reg_offset = self.project.arch.registers[cc.args[0].reg_name][
                    0
                ]
                size_arg_reg_size = cc.args[0].size
            else:
                size_arg_reg_offset = self.project.arch.registers["rdi"][0]
                size_arg_reg_size = word_size
            size = (
                state.register_definitions.load(size_arg_reg_offset, size_arg_reg_size)
                .values[0]
                .pop()
                ._model_concrete.value
            )

            # None since we do not know it's class yet, it a possible this pointer
            self.possible_objects_dict[self.max_addr] = PossibleObject(
                size, self.max_addr
            )

            # assigning eax a concrete address to track the possible this pointer
            if cc is not None:
                ret_val_reg_offset = self.project.arch.registers[
                    cc.return_val.reg_name
                ][0]
                ret_val_reg_size = cc.return_val.size
            else:
                ret_val_reg_offset = self.project.arch.registers["rax"][0]
                ret_val_reg_size = word_size
            atom = Register(ret_val_reg_offset, ret_val_reg_size)
            state.kill_and_add_definition(
                atom,
                codeloc,
                MultiValues(
                    offset_to_values={
                        0: {
                            claripy.BVV(
                                self.max_addr, word_size * state.arch.byte_width
                            )
                        }
                    }
                ),
            )
            # setting the values pointed by rax to zero
            memory_location = MemoryLocation(self.max_addr, size)
            offset_to_values = {}

            for offset in range(0, size, word_size):
                offset_to_values[offset] = {
                    claripy.BVV(0, word_size * state.arch.byte_width)
                }
            state.kill_and_add_definition(
                memory_location, codeloc, MultiValues(offset_to_values=offset_to_values)
            )
            self.max_addr += size

        elif "ctor" in self.project.kb.functions[function_address].demangled_name:
            # check if rdi has a possible this pointer/ object address, if so then we can assign this object this class
            # also if the func is a constructor(not stripped binaries)
            for addr, possible_object in self.possible_objects_dict.items():
                obj_addr = (
                    state.register_definitions.load(
                        72, state.arch.bits // state.arch.byte_width
                    )
                    .values[0]
                    .pop()
                    ._model_concrete.value
                )
                if addr == obj_addr:
                    col_ind = self.project.kb.functions[
                        function_address
                    ].demangled_name.rfind("::")
                    class_name = self.project.kb.functions[
                        function_address
                    ].demangled_name[:col_ind]
                    possible_object.class_name = class_name
        executed_rda = True
        return executed_rda, state, visited_blocks, dep_graph

class StaticObjectFinder(Analysis):
    """
    This analysis tries to find objects on the heap based on calls to new(), and subsequent calls to constructors with
     the 'this' pointer
    """

    def __init__(self):
        vtable_analysis = self.project.analyses.VtableFinder()
        self.vtables_list = vtable_analysis.vtables_list
        self.possible_objects = {}

        # for stripped binaries
        # This is a mapping between the constructors and the objects that use them
        self.possible_constructors = defaultdict(list)

        self._analyze()

    def _analyze(self):
        if "CFGFast" not in self.project.kb.cfgs:
            self.project.analyses.CFGFast(cross_references=True)
        all_functions = self.project.kb.functions
        # this is the addr where all the this pointers returned by new() will be pointing to
        max_addr = self.project.loader.main_object.max_addr + 8
        new_func_addr = self.project.kb.functions["_Znwm"].addr
        for func in all_functions:
            cc = self.project.kb.functions[func].calling_convention
            word_size = self.project.arch.bits // self.project.arch.byte_width
            # the map fo this ptrs
            newhandler = NewFunctionHandler(
                max_addr=max_addr, new_func_addr=new_func_addr, project=self.project
            )
            max_addr = newhandler.max_addr
            # this performs RDA as well as mark possible object instances for non stripped binaries
            rd = self.project.analyses.ReachingDefinitions(
                all_functions[func], observe_all=True, function_handler=newhandler
            )
            for addr, pos_obj in newhandler.possible_objects_dict.items():
                self.possible_objects[addr] = pos_obj

            # for stripped binary we check if the first function called after new(),
            # is passed the this pointer(returned by new)...
            # if so then we say that it is possibly a constructor
            for node in all_functions[func].graph.nodes():
                if all_functions[func].get_call_target(node.addr) == new_func_addr:
                    ret_node_addr = all_functions[func].get_call_return(node.addr)
                    call_after_new_addr = all_functions[func].get_call_target(
                        ret_node_addr
                    )
                    rd_before_node = rd.get_reaching_definitions_by_node(
                        ret_node_addr, OP_BEFORE
                    )

                    if cc is not None:
                        ret_val_reg_offset = self.project.arch.registers[
                            cc.return_val.reg_name
                        ][0]
                        ret_val_reg_size = cc.return_val.size
                    else:
                        ret_val_reg_offset = self.project.arch.registers["rax"][0]
                        ret_val_reg_size = word_size
                    addr_of_new_obj = (
                        rd_before_node.register_definitions.load(
                            ret_val_reg_offset, ret_val_reg_size
                        )
                        .values[0]
                        .pop()
                        ._model_concrete.value
                    )

                    rd_after_node = rd.get_reaching_definitions_by_node(
                        ret_node_addr, OP_AFTER
                    )

                    if cc is not None:
                        this_ptr_reg_offset = self.project.arch.registers[
                            cc.args[0].reg_name
                        ][0]
                        this_ptr_reg_size = cc.args[0].size
                    else:
                        this_ptr_reg_offset = self.project.arch.registers["rdi"][0]
                        this_ptr_reg_size = word_size
                    addr_in_rdi = (
                        rd_after_node.register_definitions.load(
                            this_ptr_reg_offset, this_ptr_reg_size
                        )
                        .values[0]
                        .pop()
                        ._model_concrete.value
                    )

                    if addr_of_new_obj == addr_in_rdi:
                        self.possible_constructors[call_after_new_addr].append(
                            self.possible_objects[addr_of_new_obj]
                        )


AnalysesHub.register_default("StaticObjectFinder", StaticObjectFinder)



from .coalesce_same_cascading_ifs import CoalesceSameCascadingIfs
from .constant_derefs import ConstantDereferences
from .extended_byte_and_mask import ExtendedByteAndMask
from .remove_empty_if_body import RemoveEmptyIfBody
from .remove_redundant_ite_branch import RemoveRedundantITEBranches
from .single_bit_xor import SingleBitXor
from .a_sub_a_sub_n import ASubASubN
from .conv_a_sub0_shr_and import ConvASub0ShrAnd
from .eager_eval import EagerEvaluation
from .one_sub_bool import OneSubBool
from .bool_expr_xor_1 import BoolExprXor1
from .remove_redundant_bitmasks import RemoveRedundantBitmasks
from .remove_redundant_nots import RemoveRedundantNots
from .remove_redundant_shifts import RemoveRedundantShifts
from .simplify_pc_relative_loads import SimplifyPcRelativeLoads
from .basepointeroffset_add_n import BasePointerOffsetAddN
from .basepointeroffset_and_mask import BasePointerOffsetAndMask
from .remove_redundant_conversions import RemoveRedundantConversions
from .remove_cascading_conversions import RemoveCascadingConversions
from .conv_shl_shr import ConvShlShr

from .base import PeepholeOptimizationExprBase, PeepholeOptimizationStmtBase


STMT_OPTS = [ ]
EXPR_OPTS = [ ]

_g = globals().copy()
for v in _g.values():
    if (isinstance(v, type)
        and issubclass(v, PeepholeOptimizationExprBase)
        and v is not PeepholeOptimizationExprBase
    ):
        EXPR_OPTS.append(v)

    if (isinstance(v, type)
        and issubclass(v, PeepholeOptimizationStmtBase)
        and v is not PeepholeOptimizationStmtBase
    ):
        STMT_OPTS.append(v)

_g = None



import logging
import threading

from angr.errors import AngrError
from .engine import SuccessorsMixin
from ..errors import SimConcreteRegisterError

l = logging.getLogger("angr.engines.concrete")
#l.setLevel(logging.DEBUG)

try:
    from angr_targets.concrete import ConcreteTarget
except ImportError:
    ConcreteTarget = None


class SimEngineConcrete(SuccessorsMixin):
    """
    Concrete execution using a concrete target provided by the user.
    """
    def __init__(self, project):
        if not ConcreteTarget:
            l.critical("Error, can't find angr_target project")
            raise AngrError

        l.info("Initializing SimEngineConcrete with ConcreteTarget provided.")
        super(SimEngineConcrete, self).__init__()
        self.project = project
        if isinstance(self.project.concrete_target, ConcreteTarget) and \
                self.check_concrete_target_methods(self.project.concrete_target):

            self.target = self.project.concrete_target
        else:
            l.warning("Error, you must provide an instance of a ConcreteTarget to initialize a SimEngineConcrete.")
            self.target = None
            raise NotImplementedError

        self.segment_registers_already_init = False

    def process_successors(self, successors, extra_stop_points=None, memory_concretize=None,
                           register_concretize=None, timeout=0, *args, **kwargs):

        new_state = self.state
        # setup the concrete process and resume the execution
        self.to_engine(new_state, extra_stop_points, memory_concretize, register_concretize, timeout)

        # sync angr with the current state of the concrete process using
        # the state plugin
        new_state.concrete.sync()

        successors.engine = "SimEngineConcrete"
        successors.sort = "SimEngineConcrete"
        successors.add_successor(new_state, new_state.ip, new_state.solver.true, new_state.unicorn.jumpkind)
        successors.description = "Concrete Successors"
        successors.processed = True

    def to_engine(self, state, extra_stop_points, memory_concretize, register_concretize, timeout):
        """
        Handle the concrete execution of the process
        This method takes care of:
        1- Set the breakpoints on the addresses provided by the user
        2- Concretize the symbolic variables and perform the write inside the concrete process
        3- Continue the program execution.

        :param state:               The state with which to execute
        :param extra_stop_points:   list of a addresses where to stop the concrete execution and return to the
                                    simulated one
        :param memory_concretize:   list of tuples (address, symbolic variable) that are going to be written
                                    in the concrete process memory.
        :param register_concretize:  list of tuples (reg_name, symbolic variable) that are going to be written
        :param timeout:             how long we should wait the concrete target to reach the breakpoint
        :return: None
        """

        state.globals["symbion_timeout"] = False
        extra_stop_points = [] if extra_stop_points is None else extra_stop_points

        l.debug("Entering in SimEngineConcrete: simulated address %#x concrete address %#x stop points %s",
                state.addr, self.target.read_register("pc"), map(hex, extra_stop_points))

        if memory_concretize:
            l.debug("SimEngineConcrete is concretizing memory variables before resuming the concrete process")

            for sym_var in memory_concretize:
                sym_var_address = state.solver.eval(sym_var[0])
                sym_var_value = state.solver.eval(sym_var[1], cast_to=bytes)
                l.debug("Concretize memory at address %#x with value %s", sym_var_address, str(sym_var_value))
                self.target.write_memory(sym_var_address, sym_var_value, raw=True)

        if register_concretize:
            l.debug("SimEngineConcrete is concretizing registers variables before resuming the concrete process")
            for reg in register_concretize:
                register_name = reg[0]
                register_value = state.solver.eval(reg[1])
                l.debug("Concretize register %s with value %s", register_name, str(register_value))
                self.target.write_register(register_name, register_value)

        # Set breakpoint on remote target
        for stop_point in extra_stop_points:
            l.debug("Setting breakpoints at %#x", stop_point)
            self.target.set_breakpoint(stop_point, hardware=True, temporary=True)

        if timeout > 0:
            l.debug("Found timeout as option, setting it up!")

            def timeout_handler():
                self.target.stop()    # stop the concrete target now!
                state.globals["symbion_timeout"] = True  # this will end up in the timeout stash

            execution_timer = threading.Timer(timeout, timeout_handler)
            execution_timer.start()  # start the timer!

        # resuming of the concrete process, if the target won't reach the
        # breakpoint specified by the user the timeout will abort angr execution.
        l.debug("SimEngineConcrete is resuming the concrete process")
        self.target.run()
        l.debug("SimEngineConcrete has successfully resumed the process")

        if state.globals["symbion_timeout"]:
            l.critical("Timeout has been reached during resuming of concrete process")
            l.critical("This can be a bad thing ( the ConcreteTarget didn't hit your breakpoint ) or"
                       "just it will take a while.")

        # reset the alarm
        if timeout > 0:
            execution_timer.cancel()

        # removing all breakpoints set by the concrete target
        for stop_point in extra_stop_points:
            self.target.remove_breakpoint(stop_point)

        # handling the case in which the program stops at a point different than the breakpoints set
        # by the user.
        current_pc = self.target.read_register("pc")
        if current_pc not in extra_stop_points and not state.globals["symbion_timeout"]:
            l.critical("Stopped at unexpected location inside the concrete process: %#x", current_pc)
            raise AngrError

    @staticmethod
    def check_concrete_target_methods(concrete_target):
        """
        Check if the concrete target methods return the correct type of data
        :return: True if the concrete target is compliant
        """
        entry_point = concrete_target.read_register("pc")
        if not type(entry_point) is int:
            l.error("read_register result type is %s, should be <type 'int'>", (type(entry_point)))
            return False

        mem_read = concrete_target.read_memory(entry_point, 0x4)

        if not type(mem_read) is bytes:
            l.error("read_memory result type is %s, should be <type 'bytes'>", (type(mem_read)))
            return False

        try:
            concrete_target.read_register("not_existent_reg")
            l.error("read_register should raise a SimConcreteRegisterError when accessing non existent registers")
            return False

        except SimConcreteRegisterError:
            l.debug("read_register raise a SimConcreteRegisterError, ok!")

        return True
